import multiprocessing

from pytconf.config import ParamCreator, Config

CHECK_NON_ASCII = False
SKIP_COMMENTS = False
VALIDATE_ALL_LINES_SAME_NUMBER_OF_FIELDS = True
USE_ANY_FORMAT = True
SANITIZE = True
CLEAN_EDGES = True
SUB_TRAILING = True
REMOVE_NON_ASCII = True
LOWER_CASE = True
CHECK_NUM_FIELDS = True
CONVERT_TO_STRING = True
DO_GZIP = False
FILENAME_DETECT = True
DEFAULT_ENCODING = 'utf-8'
ATTACH_ENCODER = False


class ConfigInputFiles(Config):
    """
    Parameters to specify input files
    """
    input_files = ParamCreator.create_list_str(
        help_string="input files",
    )


class ConfigNumFields(Config):
    """
    Parameter to config number of fields in a TSV file
    """
    num_fields = ParamCreator.create_int_or_none(
        help_string="how many fields should the tsv have",
        default=None,
    )


class ConfigTsvReader(Config):
    """
    Parameters to configure a TSV reader object
    """
    check_non_ascii = ParamCreator.create_bool(
        help_string="check non ascii",
        default=CHECK_NON_ASCII,
    )
    validate_all_lines_same_number_of_fields = ParamCreator.create_bool(
        help_string="validate all lines same number of fields",
        default=VALIDATE_ALL_LINES_SAME_NUMBER_OF_FIELDS,
    )


class ConfigParallel(Config):
    """
    Parameters to configure how thing should run in parallel
    """
    parallel = ParamCreator.create_bool(
        help_string="do things in parallel?",
        default=False,
    )
    jobs = ParamCreator.create_int(
        help_string="how many jobs to run in parallel",
        default=multiprocessing.cpu_count(),
    )


class ConfigInputFile(Config):
    """
    Parameters to specify input file
    """
    input_file = ParamCreator.create_list_str(
        help_string="input file",
    )


class ConfigFloatingPoint(Config):
    """
    Parameters to select whether to work with floating
    point or not
    """
    floating_point = ParamCreator.create_bool(
        default=True,
        help_string="work with floating point?",
    )


class ConfigMajority(Config):
    """
    Config the parameters for the majority algorithm
    """
    input_first_column = ParamCreator.create_int(
        help_string="first column",
    )
    input_second_column = ParamCreator.create_int(
        help_string="second column",
    )
    input_multiplication_column = ParamCreator.create_int(
        help_string="multiplication column",
    )


class ConfigProgress(Config):
    """
    Parameters to control progress reporting
    """
    progress = ParamCreator.create_bool(default=True, help_string="do you want progress report")


class ConfigColumn(Config):
    """
    Parameters to select which column to work on
    """
    column = ParamCreator.create_int(
        help_string="column to use (list of numbers separated by comma)",
    )


class ConfigColumns(Config):
    """
    Parameters to select which columns to use
    """
    columns = ParamCreator.create_list_int(
        help_string="columns to use (list of numbers separated by comma)",
    )


class ConfigAggregateColumns(Config):
    """
    Parameters to select which columns to aggregate
    """
    aggregate_columns = ParamCreator.create_list_int(
        help_string="column to aggregate by (list of numbers separated by comma)",
    )


class ConfigMatchColumns(Config):
    """
    Parameters to select which columns to match by
    """
    match_columns = ParamCreator.create_list_int(
        help_string="column to match by (list of numbers separated by comma)",
    )


class ConfigOutputFile(Config):
    """
    Parameters to configure the output file
    """
    output_file = ParamCreator.create_new_file(
        help_string="output file to generate",
    )


class ConfigTree(Config):
    """
    Parameters to configure the parameters of a tree to show
    """
    roots = ParamCreator.create_list_str(
        help_string="roots to output, separated by commas",
    )
    parent_column = ParamCreator.create_int(
        help_string="parent column",
    )
    child_column = ParamCreator.create_int(
        help_string="child column",
    )


class ConfigCsvToTsv(Config):
    """
    Parameters to control the CSV to TSV conversion process
    """
    set_max = ParamCreator.create_bool(
        help_string="do you want to unset the limit on csv fields (good for large fielded csv files)",
        default=True,
    )
    replace_tabs_with_spaces = ParamCreator.create_bool(
        help_string="do you want to replace tabs with spaces?",
        default=True,
    )
    check_num_fields = ParamCreator.create_bool(
        help_string="check that all lines have the same number of fields?",
        default=CHECK_NUM_FIELDS,
    )


class ConfigBucketNumber(Config):
    """
    Parameters to configure the bucket number for a histogram
    """
    bucket_number = ParamCreator.create_int(
        help_string="what column to histogram",
        default=10,
    )


class ConfigPattern(Config):
    """
    Parameters to configure pattern of files to be created
    """
    pattern = ParamCreator.create_str(
        help_string="pattern of generated files",
        default="{key}.tsv.gz",
    )


class ConfigFixTypes(Config):
    """
    Parameters to control which fixes to apply to a TSV file.
    """
    clean_edges = ParamCreator.create_bool(
        help_string="remove space before and after",
        default=CLEAN_EDGES,
    )
    sub_trailing = ParamCreator.create_bool(
        help_string="substitute consecutive white spaces with one single space",
        default=SUB_TRAILING,
    )
    remove_non_ascii = ParamCreator.create_bool(
        help_string="remove non ascii characters",
        default=REMOVE_NON_ASCII,
    )
    lower_case = ParamCreator.create_bool(
        help_string="lower case the field",
        default=LOWER_CASE,
    )


class ConfigSampleByColumn(Config):
    """
    Parameters for the sample by column command
    """
    weight_column = ParamCreator.create_int(
        help_string="what column to sample by",
    )
    value_column = ParamCreator.create_int(
        help_string="what is the value column",
    )
    size = ParamCreator.create_int(
        help_string="what sample size do you need?",
    )
    replace = ParamCreator.create_bool(
        help_string="allow replacements?",
        default=False,
    )
    check_unique = ParamCreator.create_bool(
        help_string="check that the value_column has unique values?",
        default=True,
    )


class ConfigSampleByColumnOld(Config):
    """
    Parameters to configure the old sample by column algorithm
    """
    sample_column = ParamCreator.create_int(
        help_string="what column to sample by",
    )
    size = ParamCreator.create_int(
        help_string="what sample size do you need?",
    )
    replace = ParamCreator.create_bool(
        help_string="allow replacement",
    )
    hits_mode = ParamCreator.create_bool(
        help_string="sample size is hits",
    )


class ConfigJoin(Config):
    """
    Parameters to configure a TSV join operation
    """
    hash_file = ParamCreator.create_existing_file(
        help_string="hash file",
    )
    hash_key_column = ParamCreator.create_int(
        help_string="column to match on in the hash file",
    )
    hash_value_column = ParamCreator.create_int(
        help_string="value to get from the hash file",
    )
    input_key_column = ParamCreator.create_int(
        help_string="column to match on in the input file",
    )
    output_insert_column = ParamCreator.create_int(
        help_string="column to insert at in the first file to create the output file",
    )
    output_add_unknown = ParamCreator.create_bool(
        help_string="add UNKNOWN when there is no match or drop",
    )


class ConfigSplit(Config):
    """
    Parameters to configure the split by column algorithm
    """
    pattern = ParamCreator.create_str(
        help_string="pattern of intermediate generated files",
        default="{key}_{i:04d}.tsv.gz",
    )
    final_pattern = ParamCreator.create_str(
        help_string="pattern of final generated files",
        default="{key}.tsv.gz",
    )


class ConfigSampleByTwoColumns(Config):
    """
    Parameters for the sample by column command
    """
    group_column = ParamCreator.create_int(
        help_string="what column to determine group by",
    )
    weight_column = ParamCreator.create_int(
        help_string="what column to determine weight by",
    )
    value_column = ParamCreator.create_int(
        help_string="what is the value column",
    )
    size = ParamCreator.create_int(
        help_string="what sample size do you need?",
    )
    replace = ParamCreator.create_bool(
        help_string="allow replacements?",
    )
    check_unique = ParamCreator.create_bool(
        help_string="check that the value_column has unique values?",
        default=True,
    )



import concurrent.futures
import csv
import logging
import os
import sys
from collections import defaultdict
from enum import Enum

import attr
import numpy
import numpy.random
import pandas
import pyanyzip.core
import pylogconf.core
import tqdm
from pytconf.config import register_endpoint, register_function_group
from typing import List, Dict, Set

from pytsv.configs import ConfigInputFiles, ConfigFloatingPoint, ConfigAggregateColumns, ConfigMatchColumns, \
    ConfigOutputFile, ConfigProgress, ConfigParallel, ConfigNumFields, ConfigTsvReader, ConfigColumns, \
    ConfigInputFile, ConfigFixTypes, ConfigColumn, ConfigBucketNumber, ConfigMajority, ConfigCsvToTsv, ConfigJoin, \
    ConfigSplit, ConfigTree, ConfigSampleByColumn, ConfigSampleByColumnOld, ConfigSampleByTwoColumns, ConfigPattern
from pytsv.core import TsvReader, TsvWriter, clean

GROUP_NAME_DEFAULT = "default"
GROUP_DESCRIPTION_DEFAULT = "all pytsv commands"


def register_group_default():
    register_function_group(
            function_group_name=GROUP_NAME_DEFAULT,
            function_group_description=GROUP_DESCRIPTION_DEFAULT,
    )


@register_endpoint(
    configs=[
        ConfigInputFiles,
        ConfigFloatingPoint,
        ConfigAggregateColumns,
        ConfigMatchColumns,
        ConfigOutputFile,
    ],
    suggest_configs=[
    ],
    group=GROUP_NAME_DEFAULT,
)
def aggregate():
    # type: () -> None
    """
    aggregate TSV files
    """
    aggregate(
        input_file_names=ConfigInputFiles.input_files,
        match_columns=ConfigMatchColumns.match_columns,
        aggregate_columns=ConfigAggregateColumns.aggregate_columns,
        output_file_name=ConfigOutputFile.output_file,
        floating_point=ConfigFloatingPoint.floating_point,
    )


class ParamsForJob:
    def __init__(self):
        self.input_file = None
        self.num_fields = None
        self.validate_all_lines_same_number_of_fields = None
        self.check_non_ascii = None
        self.progress = None


def check_file(params_for_job):
    # type: (ParamsForJob) -> bool
    print('checking [{}]...'.format(params_for_job.input_file))
    with TsvReader(filename=params_for_job.input_file,
                   num_fields=params_for_job.num_fields,
                   validate_all_lines_same_number_of_fields=params_for_job.validate_all_lines_same_number_of_fields,
                   check_non_ascii=params_for_job.check_non_ascii) as input_file_handle:
        if params_for_job.progress:
            input_file_handle = tqdm.tqdm(input_file_handle)
        for _ in input_file_handle:
            pass
    return True


@register_endpoint(
    configs=[
        ConfigProgress,
        ConfigParallel,
        ConfigNumFields,
        ConfigInputFiles,
        ConfigTsvReader,
    ],
    suggest_configs=[
    ],
    group=GROUP_NAME_DEFAULT,
)
def check():
    # type () -> None
    """
    check that every file is legal TSV
    """
    """
    TODO:
    - add ability to say how many lines are bad and print their content
    """
    if ConfigParallel.parallel:
        with concurrent.futures.ProcessPoolExecutor(max_workers=ConfigParallel.jobs) as executor:
            job_list = []
            for input_file in ConfigInputFiles.input_files:
                job = ParamsForJob()
                job.progress = ConfigProgress.progress
                job.check_non_ascii = ConfigTsvReader.check_non_ascii
                job.num_fields = ConfigNumFields.num_fields
                job.input_file = input_file
                job.validate_all_lines_same_number_of_fields = ConfigTsvReader.validate_all_lines_same_number_of_fields
                job_list.append(job)
            results = list(executor.map(check_file, job_list))
        print(results)
    for input_file in ConfigInputFiles.input_files:
        with TsvReader(
            filename=input_file,
            num_fields=ConfigNumFields.num_fields,
            validate_all_lines_same_number_of_fields=ConfigTsvReader.validate_all_lines_same_number_of_fields,
            check_non_ascii=ConfigTsvReader.check_non_ascii,
        ) as input_file_handle:
            if ConfigProgress.progress:
                input_file_handle = tqdm.tqdm(input_file_handle, desc=input_file)
            for _ in input_file_handle:
                pass


@register_endpoint(
    configs=[
        ConfigProgress,
        ConfigInputFiles,
        ConfigColumns,
    ],
    suggest_configs=[
    ],
    group=GROUP_NAME_DEFAULT,
)
def check_columns_unique():
    # type: () -> None
    """
    checks that for certain columns every value in the files is unique
    """
    dicts = [dict() for _ in range(len(ConfigColumns.columns))]
    errors = False
    for input_file in ConfigInputFiles.input_files:
        with TsvReader(
            filename=input_file,
        ) as input_file_handle:
            if ConfigProgress.progress:
                input_file_handle = tqdm.tqdm(input_file_handle, desc=input_file)
            for line_number, fields in enumerate(input_file_handle):
                for i, column in enumerate(ConfigColumns.columns):
                    value = fields[column]
                    if value in dicts[i]:
                        line = dicts[i][value]
                        print("value [{}] is duplicate on lines [{}, {}]".format(
                            value,
                            line,
                            line_number,
                        ))
                        errors = True
                    else:
                        dicts[i][value] = line_number
    assert errors is False, "found errors"


@register_endpoint(
    configs=[
        ConfigInputFile,
        ConfigOutputFile,
        ConfigProgress,
        ConfigColumns,
    ],
    suggest_configs=[
    ],
    group=GROUP_NAME_DEFAULT,
)
def clean_by_field_num():
    # type: () -> None
    """
    remove lines from a TSV file that do not have the right number of columns
    """
    with TsvReader(
        filename=ConfigInputFile.input_file,
        validate_all_lines_same_number_of_fields=False
    ) as input_file_handle:
        with TsvWriter(filename=ConfigOutputFile.output_file) as output_file_handle:
            if ConfigProgress.progress:
                input_file_handle = tqdm.tqdm(input_file_handle, desc=ConfigInputFile.input_file)
            for fields in input_file_handle:
                if len(fields) == ConfigColumns.columns:
                    output_file_handle.write(fields)


@register_endpoint(
    configs=[
        ConfigInputFile,
        ConfigOutputFile,
        ConfigColumns,
        ConfigProgress,
    ],
    suggest_configs=[
    ],
    group=GROUP_NAME_DEFAULT,
)
def cut():
    # type: () -> None
    """ cut fields from a TSV file """
    with TsvReader(filename=ConfigInputFile.input_file) as input_file_handle:
        with TsvWriter(filename=ConfigOutputFile.output_file) as output_file_handle:
            if ConfigProgress.progress:
                input_file_handle = tqdm.tqdm(input_file_handle)
            for fields in input_file_handle:
                out_fields = [fields[x] for x in ConfigColumns.columns]
                output_file_handle.write(out_fields)


@register_endpoint(
    configs=[
        ConfigInputFile,
        ConfigOutputFile,
        ConfigProgress,
        ConfigColumns,
    ],
    suggest_configs=[
    ],
    group=GROUP_NAME_DEFAULT,
)
def drop_duplicates_by_columns():
    # type: () -> None
    """
    fix a TSV file assuming that bad characters or tabs have been left in one column of it
    """
    with TsvReader(filename=ConfigInputFile.input_file) as input_file_handle:
        if ConfigProgress.progress:
            input_file_handle = tqdm.tqdm(input_file_handle)
        saw = set()
        with TsvWriter(filename=ConfigOutputFile.output_file) as output_file_handle:
            for fields in input_file_handle:  # type: List[str]
                match = frozenset([fields[match_column] for match_column in ConfigColumns.columns])
                if match not in saw:
                    saw.add(match)
                    output_file_handle.write(fields)


@register_endpoint(
    configs=[
        ConfigInputFile,
        ConfigProgress,
        ConfigOutputFile,
        ConfigColumns,
        ConfigTsvReader,
        ConfigFixTypes,
    ],
    suggest_configs=[
    ],
    group=GROUP_NAME_DEFAULT,
)
def fix_columns():
    # type: () -> None
    """
    fix a TSV file assuming that bad characters or tabs have been left in one column of it
    """
    # We need to read the input file WITHOUT assuming that it hasn't problems
    with TsvReader(
            filename=ConfigInputFile.input_file,
            check_non_ascii=ConfigTsvReader.check_non_ascii,
    ) as input_file_handle:
        if ConfigProgress.progress:
            input_file_handle = tqdm.tqdm(input_file_handle)
        with TsvWriter(filename=ConfigOutputFile.output_file) as output_file_handle:
            for fields in input_file_handle:  # type: List[str]
                for fix_column in ConfigColumns.columns:
                    fields[fix_column] = clean(
                        text=fields[fix_column],
                        clean_edges=ConfigFixTypes.clean_edges,
                        sub_trailing=ConfigFixTypes.sub_trailing,
                        remove_non_ascii=ConfigFixTypes.remove_non_ascii,
                        lower_case=ConfigFixTypes.lower_case,
                    )
                output_file_handle.write(fields)


@register_endpoint(
    configs=[
        ConfigInputFile,
        ConfigOutputFile,
        ConfigColumn,
        ConfigBucketNumber,
    ],
    suggest_configs=[
    ],
    group=GROUP_NAME_DEFAULT,
)
def histogram_by_column():
    # type: () -> None
    """ Create a histogram from a field in a TSV file """
    a = []
    total = 0
    with TsvReader(ConfigInputFile.input_file) as input_handle:
        for fields in input_handle:
            a.append(float(fields[ConfigColumn.column]))
            total += 1
    count_in_bucket, bucket_edges = numpy.histogram(a, bins=ConfigBucketNumber.bucket_number)
    with TsvWriter(ConfigOutputFile.output_file) as output_handle:
        current_sum = 0
        for i, count in enumerate(count_in_bucket):
            current_sum += count
            edge_from = bucket_edges[i]
            edge_to = bucket_edges[i+1]
            output_handle.write([
                str(edge_from),
                str(edge_to),
                str(count),
                str(int(100.0*current_sum/total)),
            ])


@register_endpoint(
    configs=[
        ConfigInputFile,
        ConfigOutputFile,
        ConfigProgress,
        ConfigMajority,
    ],
    suggest_configs=[
    ],
    group=GROUP_NAME_DEFAULT,
)
def majority():
    # type: () -> None
    """
    reduce two columns to a majority
    """
    """
    This means that if x1 appears more
    with y2 than any other values in column Y then x1, y2 will be in the output
    and no other entry with x1 will appear
    """
    d = defaultdict(dict)  # type: Dict[Dict[str, int]]
    with TsvReader(filename=ConfigInputFile.input_file) as input_file_handle:
        if ConfigProgress.progress:
            input_file_handle = tqdm.tqdm(input_file_handle)
        for fields in input_file_handle:
            p_first = fields[ConfigMajority.input_first_column]
            p_second = fields[ConfigMajority.input_second_column]
            p_multiplication = int(fields[ConfigMajority.input_multiplication_column])
            if p_second not in d[p_first]:
                d[p_first][p_second] = 0
            d[p_first][p_second] += p_multiplication
    with TsvWriter(filename=ConfigOutputFile.output_file) as output_file_handle:
        for p_first, p_dict in d.items():
            p_second = max(p_dict.keys(), key=lambda x: p_dict[x])
            p_count = p_dict[p_second]
            output_file_handle.write([
                p_first,
                p_second,
                str(p_count),
            ])


@register_endpoint(
    configs=[
        ConfigInputFile,
        ConfigOutputFile,
        ConfigProgress,
        ConfigColumn,
    ],
    suggest_configs=[
    ],
    group=GROUP_NAME_DEFAULT,
)
def multiply():
    # type: () -> None
    """ multiply a TSV file according to column """
    with TsvReader(filename=ConfigInputFile.input_file) as input_file_handle:
        with TsvWriter(filename=ConfigOutputFile.output_file) as output_file_handle:
            if ConfigProgress.progress:
                input_file_handle = tqdm.tqdm(input_file_handle)
            for fields in input_file_handle:
                current_number = int(fields[ConfigColumn.column])
                for _ in range(current_number):
                    output_file_handle.write(fields)


@register_endpoint(
    configs=[
        ConfigProgress,
        ConfigInputFiles,
    ],
    suggest_configs=[
    ],
    group=GROUP_NAME_DEFAULT,
)
def read():
    # type: () -> None
    """
    read TSV files as plainly as possible
    """
    for input_file in ConfigInputFiles.input_files:
        with TsvReader(filename=input_file) as input_file_handle:
            if ConfigProgress.progress:
                input_file_handle = tqdm.tqdm(input_file_handle, desc=input_file)
            for _ in input_file_handle:
                pass


@register_endpoint(
    configs=[
        ConfigProgress,
        ConfigColumns,
        ConfigInputFile,
        ConfigOutputFile,
    ],
    suggest_configs=[
    ],
    group=GROUP_NAME_DEFAULT,
)
def remove_quotes():
    # type: () -> None
    """ removed quotes from fields """
    with TsvReader(filename=ConfigInputFile.input_file) as input_file_handle:
        with TsvWriter(filename=ConfigOutputFile.output_file) as output_file_handle:
            if ConfigProgress.progress:
                input_file_handle = tqdm.tqdm(input_file_handle)
            for fields in input_file_handle:  # type: List[str]
                for i in ConfigColumns.columns:
                    if fields[i].startswith("\"") and fields[i].endswith("\"") and len(fields[i]) > 1:
                        fields[i] = fields[i][1:-1]
                output_file_handle.write(fields)


@register_endpoint(
    configs=[
        ConfigInputFile,
        ConfigOutputFile,
        ConfigCsvToTsv,
    ],
    suggest_configs=[
    ],
    group=GROUP_NAME_DEFAULT,
)
def csv_to_tsv():
    # type: () -> None
    """ convert a CSV to a TSV file """
    if ConfigCsvToTsv.set_max:
        csv.field_size_limit(sys.maxsize)
    with pyanyzip.core.open(ConfigInputFile.input_file, "rt") as input_file_handle:
        csv_reader = csv.reader(input_file_handle)
        with TsvWriter(
            filename=ConfigOutputFile.output_file,
            check_num_fields=ConfigCsvToTsv.check_num_fields,
        ) as output_file_handle:
            for row in csv_reader:  # type: List[str]
                if ConfigCsvToTsv.replace_tabs_with_spaces:
                    for i, item in enumerate(row):
                        row[i] = row[i].replace("\t", " ")
                output_file_handle.write(row)


class MyEventTypes(Enum):
    key_not_found = 0
    key_found = 1
    unknown_added = 2


@register_endpoint(
    configs=[
        ConfigInputFile,
        ConfigOutputFile,
        ConfigProgress,
        ConfigJoin,
    ],
    suggest_configs=[
    ],
    group=GROUP_NAME_DEFAULT,
)
def join():
    # type: () -> None
    """
    join two TSV files by column
    """
    d = dict()
    event_found = 0
    event_unknown_added = 0
    event_discarded = 0
    with TsvReader(ConfigJoin.hash_file) as hash_file_handle:
        if ConfigProgress.progress:
            hash_file_handle = tqdm.tqdm(hash_file_handle, desc="reading hash")
        for fields in hash_file_handle:
            key = fields[ConfigJoin.hash_key_column]
            value = fields[ConfigJoin.hash_value_column]
            d[key] = value
    with TsvReader(ConfigInputFile.input_file) as input_file_handle, \
            TsvWriter(ConfigOutputFile.output_file) as output_file_handle:
        if ConfigProgress.progress:
            input_file_handle = tqdm.tqdm(input_file_handle, desc="reading input and writing output")
        for fields in input_file_handle:
            key = fields[ConfigJoin.input_key_column]
            if key in d:
                event_found += 1
                new_value = d[key]
                fields.insert(ConfigJoin.output_insert_column, new_value)
                output_file_handle.write(fields)
            else:
                if ConfigJoin.output_add_unknown:
                    event_unknown_added += 1
                    fields.insert(ConfigJoin.output_insert_column, "unknown")
                    output_file_handle.write(fields)
                else:
                    event_discarded += 1
    print("event_found {}".format(event_found))
    print("event_unknown_added {}".format(event_unknown_added))
    print("event_discarded {}".format(event_discarded))


@register_endpoint(
    configs=[
        ConfigInputFile,
        ConfigOutputFile,
        ConfigColumns,
        ConfigProgress,
    ],
    suggest_configs=[
    ],
    group=GROUP_NAME_DEFAULT,
)
def lc():
    # type: () -> None
    """ lower case some columns """
    with TsvReader(filename=ConfigInputFile.input_file) as input_file_handle:
        with TsvWriter(filename=ConfigOutputFile.output_file) as output_file_handle:
            if ConfigProgress.progress:
                input_file_handle = tqdm.tqdm(input_file_handle)
            for fields in input_file_handle:  # type: List[str]
                for i in ConfigColumns.columns:
                    fields[i] = fields[i].lower()
                output_file_handle.write(fields)


@register_endpoint(
    configs=[
        ConfigInputFile,
        ConfigColumns,
        ConfigProgress,
    ],
    suggest_configs=[
    ],
    group=GROUP_NAME_DEFAULT,
)
def sum():
    # type: () -> None
    """ sum some columns """
    sums = [0] * len(ConfigColumns.columns)
    with TsvReader(filename=ConfigInputFile.input_file) as input_file_handle:
        if ConfigProgress.progress:
            input_file_handle = tqdm.tqdm(input_file_handle)
        for fields in input_file_handle:  # type: List[str]
            for n, i in enumerate(ConfigColumns.columns):
                sums[n] += float(fields[i])
    print(sums)


@attr.attrs
class JobInfo(object):
    check_not_ascii = attr.attrib()  # type: bool
    input_file = attr.attrib()  # type: str
    serial = attr.attrib()  # type: int
    progress = attr.attrib()  # type: bool
    pattern = attr.attrib()  # type: str
    columns = attr.attrib()  # type: List[int]


@attr.attrs
class JobReturnValue(object):
    serial = attr.attrib()  # type: int
    files = attr.attrib()  # type: Dict[str, str]


def process_single_file(job_info):
    # type (JobInfo) -> JobReturnValue
    logger = logging.getLogger(__name__)
    tsv_writers_dict = dict()
    results = dict()
    with TsvReader(
            filename=job_info.input_file,
            check_non_ascii=job_info.check_non_ascii
    ) as input_file_handle:
        if job_info.progress:
            logger.info("working on [%s]" % job_info.input_file)
            input_file_handle = tqdm.tqdm(input_file_handle)
        for fields in input_file_handle:
            key = ",".join([fields[x] for x in job_info.columns])
            if key not in tsv_writers_dict:
                filename = job_info.pattern.format(key=key, i=job_info.serial)
                results[key] = filename
                output_handle = TsvWriter(filename=filename)
                tsv_writers_dict[key] = output_handle
            output_handle = tsv_writers_dict[key]
            output_handle.write(fields)
    for v in tsv_writers_dict.values():
        v.close()
    return JobReturnValue(job_info.serial, results)


@register_endpoint(
    configs=[
        ConfigColumns,
        ConfigProgress,
        ConfigInputFiles,
        ConfigParallel,
        ConfigTsvReader,
        ConfigSplit,
    ],
    suggest_configs=[
    ],
    group=GROUP_NAME_DEFAULT,
)
def split_by_columns_parallel():
    # type: () -> None
    """
    split a TSV file into many files according to some of its columns
    """
    pylogconf.core.setup()
    assert len(ConfigColumns.columns) > 0, "must provide --columns"
    job_data = [JobInfo(
        ConfigTsvReader.check_non_ascii,
        input_file,
        i,
        ConfigProgress.progress,
        ConfigSplit.pattern,
        ConfigColumns.columns,
    ) for i, input_file in enumerate(ConfigInputFiles.input_files)]
    with concurrent.futures.ProcessPoolExecutor(max_workers=ConfigParallel.jobs) as executor:
        job_return_values = list(executor.map(process_single_file, job_data))  # type: List[JobReturnValue]
    job_return_values.sort(key=lambda u: u.serial)
    for job_return_value in job_return_values:
        for key, filename in job_return_value.files.items():
            outfile = ConfigSplit.final_pattern.format(key=key)
            with open(outfile, "wb") as _:
                pass
                # with open(filename, "rb") as _:


@register_endpoint(
    configs=[
        ConfigInputFile,
        ConfigTree,
    ],
    suggest_configs=[
    ],
    group=GROUP_NAME_DEFAULT,
)
def tree():
    # type: () -> None
    """
    draw tree by two columns from a TSV file
    """
    """
    You can also see only parts of the tree
    """
    children_dict = defaultdict(set)  # type: Dict[Set]
    parents_dict = defaultdict(set)
    with TsvReader(filename=ConfigInputFile.input_file) as input_file_handle:
        for fields in input_file_handle:
            p_parent = fields[ConfigTree.parent_column]
            p_child = fields[ConfigTree.child_column]
            children_dict[p_parent].add(p_child)
            parents_dict[p_child].add(p_parent)
    # find the roots (parents that have no parents)
    if ConfigTree.roots:
        list_of_roots = ConfigTree.roots.split(',')
    else:
        list_of_roots = []
        for p_parent in children_dict.keys():
            if len(parents_dict[p_parent]) == 0:
                list_of_roots.append(p_parent)

    list_to_append = []
    first = True
    for root in list_of_roots:
        list_to_append.append((root, 0, first, ""))
        first = False

    stack = []
    stack.extend(list_to_append)
    # lets draw the tree
    while len(stack) > 0:
        name, depth, last, print_list = stack.pop()
        if last:
            special_string = u"└──"
        else:
            special_string = u"├──"
        print("{}{}".format(print_list + special_string, name))
        first = True
        list_to_append = []
        for p_child in children_dict[name]:
            if last:
                special_string = "   "
            else:
                special_string = u"│  "
            list_to_append.append((p_child, depth+1, first, print_list+special_string))
            first = False
        stack.extend(list(reversed(list_to_append)))


@register_endpoint(
    configs=[
        ConfigInputFile,
        ConfigOutputFile,
    ],
    suggest_configs=[
    ],
    group=GROUP_NAME_DEFAULT,
)
def tsv_to_csv():
    # type: () -> None
    """
    convert a TSV file to a CSV file
    """
    with open(ConfigOutputFile.output_file, "wt") as output_file_handle:
        csv_writer = csv.writer(output_file_handle)
        with TsvReader(ConfigInputFile.input_file) as input_file_handle:
            for fields in input_file_handle:
                csv_writer.writerow(fields)


@register_endpoint(
    configs=[
        ConfigInputFile,
        ConfigOutputFile,
        ConfigSampleByColumn,
    ],
    suggest_configs=[
    ],
    group=GROUP_NAME_DEFAULT,
)
def sample_by_column():
    # type: () -> None
    """
    create a weighted sample from a TSV file
    """
    """
    To run this you must supply a 'value_column' (the column
    which will be sampled) and a 'weight_column' which must
    be convertible to a floating point number.
    """
    logger = logging.getLogger(__name__)
    logger.info("reading the data")
    df = pandas.read_csv(
        ConfigInputFile.input_file,
        sep='\t',
        header=None,
    )
    if ConfigSampleByColumn.check_unique:
        logger.info("checking that the values are unique")
        unique_values_count = df[ConfigSampleByColumn.value_column].nunique()
        if unique_values_count != df.shape[0]:
            logger.error("your data is not unique in the value_column")
            logger.error("unique values {} != number of rows {}".format(unique_values_count, df.shape[0]))
            return
    logger.info("sampling")
    sample = df.sample(
        n=ConfigSampleByColumn.size,
        replace=ConfigSampleByColumn.replace,
        weights=df[ConfigSampleByColumn.weight_column],
    )
    df_result = sample[ConfigSampleByColumn.value_column].value_counts()
    logger.info("writing the output")
    df_result.to_csv(
        ConfigOutputFile.output_file,
        sep='\t',
        index=True,
    )


@register_endpoint(
    configs=[
        ConfigInputFile,
        ConfigOutputFile,
        ConfigProgress,
        ConfigSampleByColumnOld,
    ],
    suggest_configs=[
    ],
    group=GROUP_NAME_DEFAULT,
)
def sample_by_column_old():
    # type: () -> None
    """
    sample from a TSV file by a sample column
    """
    weights = []
    elements = []
    sum_weights = float(0)
    with TsvReader(ConfigInputFile.input_file) as input_handle:
        if ConfigProgress.progress:
            input_handle = tqdm.tqdm(input_handle)
        for fields in input_handle:
            elements.append(fields)
            weight = float(fields[ConfigSampleByColumnOld.sample_column])
            sum_weights += weight
            weights.append(weight)
    # the following code will only work on python3.6 because the
    # random.choices API was only introduced then
    # from random import choices
    # results = choices(lines, weights, k=size)

    # this is the same code with numpy
    weights = [w/sum_weights for w in weights]
    if ConfigSampleByColumnOld.hits_mode:
        results_dict = defaultdict(int)
        for i in range(ConfigSampleByColumnOld.size):
            current_result = numpy.random.choice(
                a=len(elements),
                replace=ConfigSampleByColumnOld.replace,
                size=1,
                p=weights,
            )
            current_result = current_result[0]
            results_dict[current_result] += 1
        with TsvWriter(ConfigOutputFile.output_file) as output_handle:
            for result, hits in results_dict.items():
                record = list(elements[result])
                record.append(hits)
                output_handle.write(record)
    else:
        results = numpy.random.choice(
            a=len(elements),
            replace=ConfigSampleByColumnOld.replace,
            size=ConfigSampleByColumnOld.size,
            p=weights,
        )
        with TsvWriter(ConfigOutputFile.output_file) as output_handle:
            for result in results:
                output_handle.write(elements[result])


@register_endpoint(
    configs=[
        ConfigInputFile,
        ConfigOutputFile,
        ConfigProgress,
        ConfigSampleByTwoColumns,
    ],
    suggest_configs=[
    ],
    group=GROUP_NAME_DEFAULT,
)
def sample_by_two_columns():
    # type: () -> None
    """
    sample from a TSV file by two columns
    """
    logger = logging.getLogger(__name__)
    logger.info("reading the data")
    df = pandas.read_csv(
        ConfigInputFile.input_file,
        sep='\t',
        header=None,
    )
    logger.info("checking that the values are unique")
    unique_values_count = df[ConfigSampleByTwoColumns.value_column].nunique()
    if ConfigSampleByTwoColumns.check_unique and unique_values_count != df.shape[0]:
        logger.error("your data is not unique in the value_column")
        logger.error("unique values {} != number of rows {}".format(unique_values_count, df.shape[0]))
        return
    logger.info("finding clusters")
    clusters = df[ConfigSampleByTwoColumns.group_column].unique()
    if ConfigProgress.progress:
        clusters = tqdm.tqdm(clusters)
    if os.path.isfile(ConfigOutputFile.output_file):
        os.unlink(ConfigOutputFile.output_file)
    for cluster in clusters:
        cluster_queries = df[df[ConfigSampleByTwoColumns.group_column] == cluster]
        sample = cluster_queries.sample(
            n=ConfigSampleByTwoColumns.size,
            replace=ConfigSampleByTwoColumns.replace,
            weights=cluster_queries[ConfigSampleByTwoColumns.weight_column],
        )
        res = sample[sample.columns[ConfigSampleByTwoColumns.value_column]].value_counts()
        res.to_csv(
            ConfigOutputFile.output_file,
            sep='\t',
            index=False,
            mode='a',
            header=None,
        )


@register_endpoint(
    configs=[
        ConfigColumns,
        ConfigProgress,
        ConfigTsvReader,
        ConfigInputFiles,
        ConfigPattern,
    ],
    suggest_configs=[
    ],
    group=GROUP_NAME_DEFAULT,
)
def split_by_columns():
    # type: () -> None
    """
    split a TSV file into many files according to some of its columns
    """
    pylogconf.core.setup()
    logger = logging.getLogger(__name__)
    assert len(ConfigColumns.columns) > 0, "must provide --columns"
    tsv_writers_dict = dict()
    for input_file in ConfigInputFiles.input_files:
        with TsvReader(filename=input_file, check_non_ascii=ConfigTsvReader.check_non_ascii) as input_file_handle:
            if ConfigProgress.progress:
                logger.info("working on [%s]" % input_file)
                input_file_handle = tqdm.tqdm(input_file_handle)
            for fields in input_file_handle:
                key = ",".join([fields[x] for x in ConfigColumns.columns])
                if key not in tsv_writers_dict:
                    filename = ConfigPattern.pattern.format(key=key)
                    output_handle = TsvWriter(filename=filename)
                    tsv_writers_dict[key] = output_handle
                output_handle = tsv_writers_dict[key]
                output_handle.write(fields)
    # close all writers
    for v in tsv_writers_dict.values():
        v.close()



import setuptools

"""
The documentation can be found at:
http://setuptools.readthedocs.io/en/latest/setuptools.html
"""
setuptools.setup(
    # the first three fields are a must according to the documentation
    name='pytsv',
    version='0.1.57',
    packages=[
        'pytsv',
        'pytsv.endpoints',
    ],
    # from here all is optional
    description='pytsv is a module to help with all things TSV',
    long_description='pytsv is a module to help with all things TSV',
    author='Mark Veltzer',
    author_email='mark.veltzer@gmail.com',
    maintainer='Mark Veltzer',
    maintainer_email='mark.veltzer@gmail.com',
    keywords=[
        'python',
        'tsv',
        'format',
        'csv',
    ],
    url='https://veltzer.github.io/pytsv',
    download_url='https://github.com/veltzer/pytsv',
    license='MIT',
    platforms=[
        'python2',
    ],
    install_requires=[
        'pytconf',
        'tqdm',
        'pyanyzip',
        'numpy',
        'futures',
        'pandas',
        'pylogconf',
        'attrs',
    ],
    classifiers=[
        'Development Status :: 4 - Beta',
        'Environment :: Console',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Topic :: Utilities',
    ],
    data_files=[
    ],
    entry_points={'console_scripts': [
        'pytsv=pytsv.endpoints.main:main',
    ]},
    python_requires='>=2.7',
)



from __future__ import print_function

import codecs
import gzip
from collections import defaultdict
from typing import Iterable, List, Tuple, Dict, IO, Iterator, Text, Union, Sequence
import itertools
import logging
import re

import pyanyzip.core
import sys

from pytsv.configs import SANITIZE, CLEAN_EDGES, SUB_TRAILING, REMOVE_NON_ASCII, LOWER_CASE, CHECK_NUM_FIELDS, \
    CONVERT_TO_STRING, FILENAME_DETECT, DO_GZIP, DEFAULT_ENCODING, ATTACH_ENCODER, USE_ANY_FORMAT, SKIP_COMMENTS, \
    CHECK_NON_ASCII, VALIDATE_ALL_LINES_SAME_NUMBER_OF_FIELDS

if sys.version_info > (3, 0):
    def stream_next(file):
        return file.readline()
else:
    def stream_next(file):
        return file.next()


def is_2():
    return sys.version_info[0] == 2


def clean(
        text,
        clean_edges=True,
        sub_trailing=True,
        remove_non_ascii=True,
        lower_case=True,
):
    # type: (str, bool, bool, bool, bool) -> str
    if sub_trailing:
        # replace all manner of whitespace (consecutive or not)
        # with s single space
        text = re.sub(r"[\r\t\n\v ]+", " ", text)
    if remove_non_ascii:
        text = text.encode('ascii', errors='ignore').decode()
    if clean_edges:
        # remove space from the left and right
        text = text.strip()
    if lower_case:
        text = text.lower()
    return text


def aggregate(
    input_file_names,
    match_columns,
    aggregate_columns,
    output_file_name,
    floating_point,
):
    # type: (Iterable[str], List[int], List[int], str, bool) -> None
    """
    This function aggregates a bunch of input files by integers.
    :param input_file_names:
    :param match_columns:
    :param aggregate_columns:
    :param output_file_name:
    :param floating_point:
    :return:
    """
    counts = dict()  # type: Dict[Tuple[str], List[Union[int, float]]]
    logger = logging.getLogger(__name__)
    for input_file_name in input_file_names:
        logger.debug("working on file [%s]", input_file_name)
        with TsvReader(filename=input_file_name) as input_handle:
            for fields in input_handle:
                match = tuple([fields[i] for i in match_columns])  # type: Tuple[str]
                if match not in counts:
                    if floating_point:
                        counts[match] = [float(0)] * len(aggregate_columns)
                    else:
                        counts[match] = [int(0)] * len(aggregate_columns)
                for i, aggregate_column in enumerate(aggregate_columns):
                    if floating_point:
                        counts[match][i] += float(fields[aggregate_column])
                    else:
                        counts[match][i] += int(fields[aggregate_column])
    # notice that we create a writer that does not check the number
    # of fields because the check requires a len(fields) to be available
    # and it is not in this case because of itertools.chain
    with TsvWriter(
        filename=output_file_name,
        check_num_fields=False,
        sanitize=False,
    ) as output_file_handle:
        for match, aggregates in counts.items():
            to_write = itertools.chain(match, aggregates)  # type: List[str]
            output_file_handle.write(to_write)


def write_data(data, output_file_name):
    # type: (List[List[str]], str) -> None
    with TsvWriter(
            filename=output_file_name,
            mode="at",
    ) as output_file_handle:
        for d in data:
            output_file_handle.write(d)


def group_by(
        input_file_names,
        group_by_columns,
        collect_columns,
        output_file_template,
):
    # type: (Iterable[str], List[int], List[int], str, bool) -> List[str]
    all_data = defaultdict(list)  # type: Dict[Tuple[str], List[List[str]]]
    limit = 10000
    logger = logging.getLogger(__name__)
    for input_file_name in input_file_names:
        logger.debug("working on file [%s]", input_file_name)
        with open(input_file_name, 'rt') as file_handle:
            for line in file_handle:
                line = line.rstrip()
                parts = line.split("\t")  # type: List[str]
                match = tuple([parts[i] for i in group_by_columns])  # type: Tuple[str]
                match = "_".join(match)
                data_to_append = [parts[i] for i in collect_columns]
                all_data[match].append(data_to_append)
                if len(all_data[match]) > limit:
                    write_data(all_data[match], output_file_template.format(match=match))
                    all_data[match] = list()
    # write the rest for the data
    for match, data in all_data.items():
        write_data(data, output_file_template.format(match=match))
    return [output_file_template.format(match=match) for match in all_data]


def is_ascii(s):
    # type: (str) -> bool
    return all(ord(c) < 128 for c in s)


class TsvWriter(object):
    def __init__(
            self,
            filename,
            mode="wt",
            throw_exceptions=False,

            sanitize=SANITIZE,
            fields_to_clean=None,
            clean_edges=CLEAN_EDGES,
            sub_trailing=SUB_TRAILING,
            remove_non_ascii=REMOVE_NON_ASCII,
            lower_case=LOWER_CASE,

            check_num_fields=CHECK_NUM_FIELDS,
            num_fields=None,
            convert_to_string=CONVERT_TO_STRING,

            do_gzip=DO_GZIP,
            filename_detect=FILENAME_DETECT,
            encoding=DEFAULT_ENCODING,
            attach_encoder=ATTACH_ENCODER,
    ):
        # type: (str, str, bool, bool, List[int], bool, bool, bool, bool, bool, int, bool, bool, bool) -> None
        if filename_detect:
            found = False
            if filename.endswith(".tsv.gz"):
                self.io = gzip.open(filename, mode=mode)  # type: IO[str]
                if is_2():
                    self.io = codecs.getwriter(encoding=encoding)(self.io)
                found = True
            if filename.endswith(".tsv"):
                self.io = open(filename, mode=mode)  # type: IO[str]
                if is_2():
                    self.io = codecs.getwriter(encoding=encoding)(self.io)
                found = True
            if not found:
                # treat as tsv
                self.io = open(filename, mode=mode)  # type: IO[str]
            # old code, be more strict
            # assert found, "file name unknown"
        else:
            if do_gzip:
                self.io = gzip.open(filename, mode=mode)  # type: IO[str]
                if is_2():
                    self.io = codecs.getwriter(encoding=encoding)(self.io)
            else:
                self.io = open(filename, mode=mode)  # type: IO[str]
        # the next branch is mainly for python 2 when the PYTHONIOENCODING
        # environment variable is not set
        if attach_encoder:
            self.io = codecs.getwriter(encoding=encoding)(self.io)
        self.throw_exceptions = throw_exceptions

        self.sanitize = sanitize
        self.fields_to_clean = fields_to_clean
        if self.fields_to_clean is None:
            self.fields_to_clean = []
        self.clean_edges = clean_edges
        self.sub_trailing = sub_trailing
        self.remove_non_ascii = remove_non_ascii
        self.lower_case = lower_case

        self.check_num_fields = check_num_fields
        self.num_fields = num_fields
        self.convert_to_string = convert_to_string

    def _sanitize(self, l):
        # type: (Sequence[str]) -> Sequence[str]
        if self.sanitize:
            r = list(l)
            for field in self.fields_to_clean:
                r[field] = clean(
                    text=r[field],
                    clean_edges=self.clean_edges,
                    sub_trailing=self.sub_trailing,
                    remove_non_ascii=self.remove_non_ascii,
                    lower_case=self.lower_case,
                )
            return r
        else:
            return l

    def _convert(self, l):
        # type: (Sequence[str]) -> Iterator[str]
        if self.convert_to_string:
            for i, t in enumerate(l):
                if type(t) in (int, float, type(None)):
                    yield str(t)
                else:
                    yield t
        else:
            for x in l:
                yield x

    def write(self, input_list):
        # type: (Sequence[str]) -> None
        sanitized_list = self._sanitize(input_list)
        if self.check_num_fields:
            if self.num_fields is None:
                self.num_fields = len(sanitized_list)
            else:
                assert len(sanitized_list) == self.num_fields, "wrong number of fields in {}".format(sanitized_list)
        print("\t".join(self._convert(sanitized_list)), file=self.io)

    def close(self):
        # type: () -> None
        self.io.close()

    def __enter__(self):
        """ method needed to be a context manager """
        return self

    def __exit__(self, itype, value, traceback):
        """ method needed to be a context manager """
        self.close()


class TsvReader:
    def __init__(
            self,
            filename,
            mode="rt",
            use_any_format=USE_ANY_FORMAT,
            validate_all_lines_same_number_of_fields=VALIDATE_ALL_LINES_SAME_NUMBER_OF_FIELDS,
            num_fields=None,
            skip_comments=SKIP_COMMENTS,
            check_non_ascii=CHECK_NON_ASCII,
            newline='\n',
    ):
        # type: (str, str, bool, bool, int, bool, bool, Union[str, None]) -> None
        if use_any_format:
            self.io = pyanyzip.core.open(name=filename, mode=mode, newline=newline)
        else:
            self.io = open(filename, mode=mode, newline=newline)
        self.validate_all_lines_same_number_of_fields = validate_all_lines_same_number_of_fields
        self.num_fields = num_fields
        self.skip_comments = skip_comments
        self.check_non_ascii = check_non_ascii

        self.line_number = -1

    def next(self):
        return self.__next__()

    def __next__(self):
        # type: () -> List[Text]
        """ method needed to be an iterator """
        self.line_number += 1
        line = stream_next(self.io)
        if not line:
            raise StopIteration
        if self.skip_comments:
            while line.startswith("#"):
                self.line_number += 1
                line = stream_next(self.io)
                if not line:
                    raise StopIteration
        line = line.rstrip('\r\n')
        if self.check_non_ascii:
            assert is_ascii(line), "non ascii characters in line [{}]".format(self.line_number)
        fields = line.split('\t')
        if self.validate_all_lines_same_number_of_fields:
            if self.num_fields is None:
                self.num_fields = len(fields)
            else:
                assert len(fields) == self.num_fields, "wrong number of fields in line number {} {} {}".format(
                    self.line_number,
                    self.num_fields,
                    len(fields),
                )
        if self.check_non_ascii:
            assert is_ascii(line)
        return fields

    def __iter__(self):
        """ method needed to be an iterator """
        return self

    def __enter__(self):
        """ method needed to be a context manager """
        return self

    def __exit__(self, itype, value, traceback):
        """ method needed to be a context manager """
        self.close()

    def close(self):
        # type: () -> None
        self.io.close()


def write_dict(filename=None, d=None):
    # type: (str, Dict[str,str]) -> None
    with TsvWriter(filename=filename, num_fields=2, fields_to_clean=[0]) as output_handle:
        for k, v in d.items():
            output_handle.write([k, v])
